/***************************************************************************\
*               Comment Mode                                                *
*                                                                           *
*       This file contains routines and commands dealing with comments      *
*       Larry Osterman                                                      *
\**************************************************************************/

/*
 * Commented and tweaked a little by K. Shane Hartman for Epsilon 6.0
 * ANSI style comments (//) added by Aaron A. Collins and Bill Minus for v6.5
 *
 * Minor corrections by Lugaru, 7/10/93: fixed commenting problem that
 * prevented compilation (how ironic!); made it use sprintf's %q instead of
 * manually (& incorrectly) quoting regex search string; made it use
 * comment_begin (preferred-comment-string variable), not comment_start
 * (regex pattern for all possible comments).
 *
 *
 * This package is still better than Lugaru's because Lugaru ignores
 * EXISTING comments (this package moves to the comment and indents it).
 * [Note: Epsilon 6.X's built-in commenter DOES recognize existing 
 *  comments.  -Lugaru Software]
 * Also, this package can comment out regions.  [True enough.  -Lugaru]
 *
 * This file implements command for manipulating comments.  It uses three
 * buffer specific variables: comment_begin, comment_end and comment_column.
 * See below for a description.  The commands implemented are:
 *
 * set-comment-column (C-x ;)  Sets the comment column to the current column,
 *    with an arg, sets it back to the default value.
 *
 * insert-comment (A-;) Creates (or moves to) a comment.  If a comment already
 *     exists, it is moved to the comment column (unless it is the first thing
 *     on the line, in which case it stays put).
 *
 * insert-comment-on-next-line (A-n) Like insert-comment, but on next line.
 *
 * insert-comment-on-previous-line (A-p) Like insert-comment, but on previous
 *    line.
 *
 * There are also commands for commenting out lines and regions at the end
 * of the file.  (A-+, A--, and C-x +).
 *
 * comment_out_next_line (A-+)
 *
 * comment_out_prev_line (A--)
 *
 * comment_out_region (C-x +)
 *
 * Additions by Aaron A. Collins, originally written by Bill Minus, 10-1-1991
 *
 * comment_line (F6) - Prepends // to the current line, and moves point down to
 *	  start of next line.  Can be hit repeatedly to comment out a block of lines.
 *
 * uncomment_line (Shift-F6) - Removes // from the current line if it's there,
 *    and moves point down to start of next line.  Can be hit repeatedly to
 *    remove // type comments from a block of lines.
 *
 * comment_region (C-x /) - Prepends // to all lines in the current region, and
 *    leaves the point at the end of the region when finished.
 *
 * uncomment_region (C-x \) - Removes // from all lines in the current region,
 *    if they are there, and leaves point at the end of the region when finished.
 *
 */
 
#include "eel.h"

delete_empty_comment_on_line()
{
    char empty_comment[FNAMELEN];
    int beginsrch,tpoint,beginmatch,endmatch;

    to_begin_line();
/*
 *      Build up a regular expression that looks like an empty comment
 *              <comment_begin>[ \t]*<comment_end>
 *
 *      Since comment_begin and comment_end can contain regular expression
 *      special characters we use sprintf's %q sequence to stick
 *      regular expression quoting characters around the comment
 *      (ie the C comment characters become: "%/%*% [ \t]*% %*%/")
 */
	 sprintf(empty_comment, "%q[ \\t]*%q\n", comment_begin, comment_end);

/*
 *      Search forward to find an empty comment
 */

    beginsrch = point;                  /* Mark the search start. */
    if (!re_search(1,empty_comment)) {  /* If we found no empty comments */
        point = beginsrch;              /* Go back to beginning of search */
        return;
    }                                   /* Otherwise.... */

/*
 *      We mark the beginning and end of the comment that we found,
 *      interchange point and the search start (to make horizontal()
 *      work properly), and see if this comment was on the
 *      same line we are on.  If so, we are in luck and we can delete
 *      the comment
 */
    endmatch = point;                   /* endmatch = end of match */
    beginmatch = matchstart;            /* beginmatch = start of match */
    point = matchstart;                 /* Point is at beginning of match */
                                        
    tpoint = beginsrch;                 /* Interchange point and beginning */
    beginsrch = point;
    point = tpoint;
    if (horizontal(beginsrch)!=-1) {    /* If we found an empty comment on 
                                         * this line */
    /*
     *  We can delete the comment on this line.
     *  We set tmpoint to hold the start location and
     *  tmatch to be the start of the match.  We then delete the
     *  comment from the buffer, stick in a \n (we deleted it
     *  from the end of the string), clean up any whitespace at the
     *  end of the line and return.
     */
        spot tmpoint = alloc_spot(),
             tmatch = alloc_spot();
        *tmpoint = point;
        *tmatch = beginmatch;
        delete(beginmatch,endmatch);    /* Kill the comment */
        point = *tmatch;
        insert('\n');                   /* Insert newline deleted in search */
        point = *tmatch;                /* Update point to before \n */
        delete_horizontal_space();      /* Kill trailing whitespace. */
        point = *tmpoint;               /* go to before deleted comment. */
        free_spot(tmpoint);
        free_spot(tmatch);
    }
													 /* Just clean up, there were */
													 /* no empty comments on this line */
}

/*      insert_comment_on_this_line - insert a comment.
 *
 *      int insert_comment_on_this_line(eol)
 *      
 *              eol - column containing the end of the current line.
 *      returns:
 *              the column of the start of the comment
 */

int insert_comment_on_this_line(eol)
int eol;
{
    int comcol;
/*
 *      If line is longer than the comment column, stick 
 *      the comment at end of the line.
 */
    if (eol > comment_column) {
        to_end_line();
        insert(' ');
        comcol = eol + 1;
    }
    else {
/*
 *      Otherwise, to to the comment column and stick it there
 */
        my_to_column(comment_column);
        comcol = comment_column;
    }
    stuff(comment_begin);
    if (comment_end) {
        stuff(comment_end);
    }
    return comcol;
}

/*
 * insert-comment (A-;) Creates (or moves to) a comment.  If a comment already
 *     exists, it is moved to the comment column (unless it is the first thing
 *     on the line, in which case it stays put).
 */
command insert_comment() on reg_tab[ALT(';')] 
{                                       
    int eol,                            
        beglin,
        tpoint,
        comcol,
        comment_begin_len = strlen(comment_begin);
/*
 *      Compute the location of the beginning of the line 
 */
    comcol = current_column(),
    to_end_line();
    eol = current_column();             /* eol = column of end of line */
    to_begin_line();
    beglin = point;                     /* beglin = beginning of line. */
/*                                      
 *      Find if there is a comment on this line
 *      We do this by searching forward from the current position
 *      for a comment beginning character.  If we find one,
 *      we update point to the beginning of the comment, if not,
 *      point is set at the end of the line
 */
    {
      int len = strlen (comment_begin);
      char *s = malloc (len + 1);
      int i;
      
      for (i = 0; i < len; i++)
        if ((s[i] = comment_begin[i]) == ' ') break;
      s[i] = 0;
      
      if (!search(1,s)) {               /* Non zero if we found a comment */
        point = beglin;                 /* Go to beginning of line, */
        comcol = insert_comment_on_this_line(eol); /* stick a comment here */
        my_to_column(comcol+comment_begin_len); /* Stick point at start */
        free (s);
        return;                         /* and return to caller. */
      }
      free (s);
    }
    point = matchstart;
    tpoint = beglin;
    beglin = point;
    point = tpoint;                     /* Swap start and end of search */
    /*
     * If newline before comment, horizontal will return -1.
     */
    if (horizontal(beglin)==-1)
        comcol = insert_comment_on_this_line(eol);
    else {
        /*
         * There is a comment on this line.  point is set to the
         * beginning of the comment string.  Move the string
         * to the comment column
         */
        point = beglin;                 /* Update point to start of comment */

        if (current_column()) {         /* If not beginning of line, */
                                        /* adjust the comment. */
            delete_horizontal_space();  /* Remove whitespace */
            eol = current_column();
            comcol=(eol>=comment_column)? eol + 1 : comment_column;
            to_column(comcol);
        }
    }
    my_to_column(comcol+comment_begin_len);
}                                       
                                        
/*  
 *      my_to_column(col) - go to this column absolutly 
 */

my_to_column(columnum)
int columnum;
{
    int eolcol;
    move_to_column(columnum);           /* Move to specified column */
    if ((eolcol = current_column()) < columnum) {
        insert_to_column(eolcol,columnum);
        to_end_line();                  /* Travel to the end of the line */
    }
}
                                        

/* insert-comment-on-next-line (A-n) Like insert-comment, but on next line. */
command insert_comment_on_next_line() on reg_tab[ALT('n')]
{
    delete_empty_comment_on_line();
    nl_forward();
    insert_comment();
}

/*
 * insert-comment-on-previous-line (A-p) Like insert-comment, but on previous
 *    line.
 */
command insert_comment_on_previous_line() on reg_tab[ALT('p')]
{
    delete_empty_comment_on_line();
    nl_reverse();
    insert_comment();
}

/*
 *
 *      comment_out_line - Comment out the current line.
 *
 *      Comments out the current line, leaving point on the start of
 *      the next line.
 */
command comment_out_next_line() on reg_tab[ALT('+')]
{
    to_end_line();
    point++;
    comment_out_line();
}

command comment_out_prev_line() on reg_tab[ALT('-')]
{
    to_begin_line();
    point--;
    comment_out_line();
}

command comment_out_line()
{
    spot oldpoint = alloc_spot();
    *oldpoint = point;
    to_begin_line();                    /* Move to beginning of line */
    stuff(comment_begin);
    to_end_line();
    stuff(comment_end);
    point = *oldpoint;
    free_spot(oldpoint);
}

command comment_out_region() on cx_tab['+']
{
    spot oldpoint;
    if (point==mark)
        return;
    oldpoint = alloc_spot();            /* allocate a spot for current point*/
    *oldpoint = point;                  /* keep track of where we started */
    if (point < mark) {
        while (point<mark) {
            comment_out_line();
            point++;
        }
    } else {
        while (point>mark) {
            comment_out_line();
            to_begin_line();
            point--;
        }
    }
    point = *oldpoint;
    free_spot(oldpoint);
    return;
}

/* Additions by Aaron A. Collins, originally written by Bill Minus, 10-1-1991 */

command comment_line() on reg_tab[FKEY(6)]
{												/* add ansi start of comment */
	if (nl_reverse())							/* to start of line */
		point++;
	insert('/');								/* add ansi start of comment */
	insert('/');								/* add ansi start of comment */
	nl_forward();								/* to next line */
}

command uncomment_line() on reg_tab[FSHIFT(6)]
{												/* remove ansi start of comment */
	if (nl_reverse())							/* to start of line */
		point++;
	if (curchar() == '/') {						/* if line starts with comment */
		point++;
		if (curchar() == '/')					/* make sure it is an ansi start of comment */
			delete(point - 1, point + 1);		/* delete ansi start of comment */
	}
	nl_forward();								/* to next line */
}

command comment_region() on cx_tab['/']
{												/* add ansi start of comment */
	int temp, start_point;						/* in region */
	int noflip;
	
	noflip = 0;
	if (point > mark) {							/* point after mark? */
		temp = mark;
		mark = point;
		point = temp;
	}
	else {
		noflip = 1;
		start_point = point;
	}
	temp = point;								/* save old point */
	while (point < mark) {
		comment_line();
	}
	if (noflip)
		point = start_point;
	else mark = temp;
	say("Done.");
}

command uncomment_region() on cx_tab['\\']
{												/* remove ansi start of comment */
	int temp, start_point;						/* in region */
	int noflip;
	
	noflip = 0;
	if (point > mark) {							/* point after mark? */
		temp = mark;
		mark = point;
		point = temp;
	}
	else {
		noflip = 1;
		start_point = point;
	}
	temp = point;								/* save old point */
	while (point < mark) {
		uncomment_line();
	}
	if (noflip)
		point = start_point;
	else
		mark = temp;
	say("Done.");
}

